from __future__ import annotations

import torch
import torchio as tio
from hydra.utils import instantiate
from transformers import BertTokenizer
from transformers.tokenization_utils_base import BatchEncoding

from .checkpoint import load_processor_config
from .types import TypeImage
from .types import TypeImages
from .types import TypeImagesTensor
from .types import TypePath
from .types import TypeStringOrStrings
from .types import TypeTextAttentionMask
from .types import TypeTextTokenIds


def get_processor(*, image_only: bool = False, **kwargs) -> Processor:
    overrides = []
    for key, value in kwargs.items():
        overrides.append(f"{key}={value}")
    config = load_processor_config(overrides=overrides)
    if image_only:
        config.tokenizer = None
    return instantiate(config)


class Processor:
    def __init__(
        self,
        image_transform: tio.Transform,
        tokenizer: BertTokenizer,
    ):
        self._image_transform = image_transform
        self._text_tokenizer = tokenizer

    def __repr__(self) -> str:
        lines = [
            f"{self.__class__.__name__}(",
            f"  image_transform={self._image_transform},",
            f"  text_tokenizer={self._text_tokenizer},",
            ")",
        ]
        return "\n".join(lines)

    def process_images(
        self,
        inputs: TypePath | TypeImage | list[TypePath | TypeImage],
    ) -> TypeImages:
        if not isinstance(inputs, list):
            inputs = [inputs]
        images = []
        for image_or_path in inputs:
            is_image = isinstance(image_or_path, tio.ScalarImage)
            if is_image:
                image = image_or_path
            else:
                path = image_or_path
                image = tio.ScalarImage(path)
            images.append(image)
        images = [self._image_transform(image) for image in images]
        return images

    def to_images_batch(
        self,
        images: TypeImages,
    ) -> TypeImagesTensor:
        if not isinstance(images, list):
            msg = f"Expected images to be a list, got {type(images)}"
            raise TypeError(msg)
        images_tensor = torch.stack([image.data for image in images])
        return images_tensor

    def process_text(
        self,
        text: TypeStringOrStrings,
        **kwargs,
    ) -> tuple[TypeTextTokenIds, TypeTextAttentionMask]:
        encodings: BatchEncoding = self._text_tokenizer(
            text,
            max_length=512,  # TODO
            padding="max_length",
            truncation=True,
            return_tensors="pt",
            **kwargs,
        )
        token_ids: TypeTextTokenIds = encodings["input_ids"]  # type: ignore[reportAssignmentType]
        attention_mask: TypeTextAttentionMask = encodings["attention_mask"]  # type: ignore[reportAssignmentType]
        return token_ids, attention_mask



import torch

from .model.multimodal import Model
from .processor import Processor
from .types import TypeImage
from .types import TypePath
from .types import TypeScores


class ZeroShotImageClassificationPipeline:
    def __init__(
        self,
        model: Model,
        processor: Processor,
        batch_size: int = 1,  # TODO
        num_workers: int = 0,  # TODO
    ):
        self._model = model.eval()
        self._processor = processor

    @property
    def device(self) -> torch.device:
        return self._model.device

    # TODO: add support for multiple prompts per class
    @torch.no_grad()
    def __call__(
        self,
        images: TypePath | TypeImage | list[TypePath | TypeImage],
        prompts: str | list[str],
    ) -> TypeScores | list[TypeScores]:
        preprocessed_images = self._processor.process_images(images)
        images_batch = self._processor.to_images_batch(preprocessed_images)

        text_token_ids, text_attention_mask = self._processor.process_text(prompts)

        image_embeddings_batch = self._model.encode_image(
            images_batch,
            project=True,
            pool=True,
            normalize=True,
        )
        text_embeddings_batch = self._model.encode_text(
            text_token_ids,
            text_attention_mask,
        )

        probabilities = self._model.classify(
            image_embeddings_batch,
            text_embeddings_batch,
        )

        if not isinstance(prompts, list):
            prompts = [prompts]

        all_results = []
        for image_probabilities in probabilities:
            image_results = []
            for prompt, score in zip(prompts, image_probabilities, strict=True):
                image_results.append({"score": score.item(), "label": prompt})
            all_results.append(image_results)

        if len(images_batch) == 1 and not isinstance(images, list):
            all_results = all_results[0]
        return all_results



from __future__ import annotations

import logging
from pathlib import Path

import torch
from accelerate import init_empty_weights
from accelerate import load_checkpoint_and_dispatch
from einops import rearrange
from hydra.utils import instantiate
from safetensors.torch import load_model
from safetensors.torch import save_model
from torch import nn
from transformers.utils.logging import get_verbosity
from transformers.utils.logging import set_verbosity
from transformers.utils.logging import set_verbosity_error

from ..checkpoint import download_weights
from ..checkpoint import load_model_config
from ..types import TypeImageEmbeddings
from ..types import TypeImagesTensor
from ..types import TypeIntOrTripletInt
from ..types import TypePatchEmbeddings
from ..types import TypePatchLogits
from ..types import TypePatchProbabilities
from ..types import TypePath
from ..types import TypePooledEmbeddings
from ..types import TypePooledLogits
from ..types import TypePooledProbabilities
from ..types import TypeTextAttentionMask
from ..types import TypeTextEmbeddings
from ..types import TypeTextTokenIds
from .base import ModelMixin
from .image import ImageEncoder
from .text import TextEncoder


def get_model(
    checkpoint_path: TypePath | None = None,
    *,
    pretrained: bool = True,
    image_only: bool = False,
    **kwargs,
) -> Model:
    if pretrained and checkpoint_path is None:
        checkpoint_path = download_weights()

    overrides = []
    for key, value in kwargs.items():
        overrides.append(f"{key}={value}")
    config = load_model_config(overrides=overrides)

    if image_only:
        config.text_encoder = None

    transformers_verbosity = get_verbosity()
    set_verbosity_error()
    if checkpoint_path is None:
        model = instantiate(config)
    else:
        with init_empty_weights():
            model = instantiate(config)
        accelerate_logger = logging.getLogger("accelerate.utils.modeling")
        old_level = accelerate_logger.getEffectiveLevel()
        accelerate_logger.setLevel(logging.ERROR)
        model = load_checkpoint_and_dispatch(model, str(checkpoint_path))
        accelerate_logger.setLevel(old_level)
    set_verbosity(transformers_verbosity)

    assert isinstance(model, Model)
    return model.eval()


class Model(nn.Module, ModelMixin):
    def __init__(
        self,
        image_encoder: ImageEncoder,
        text_encoder: TextEncoder,
        temperature: float = 1,
    ):
        super().__init__()
        self.image_encoder = image_encoder
        self.text_encoder = text_encoder
        self.register_buffer("softmax_temperature", torch.tensor(temperature))

    @property
    def patch_size(self) -> tuple[int, int, int]:
        return self.image_encoder.patch_size

    def encode_image(
        self,
        images: TypeImagesTensor,
        *,
        pool: bool = False,
        project: bool = False,
        normalize: bool = False,
        window_size: TypeIntOrTripletInt | None = None,
    ) -> TypeImageEmbeddings:
        return self.image_encoder(
            images,
            project=project,
            pool=pool,
            normalize=normalize,
            window_size=window_size,
        )

    def encode_text(
        self,
        token_ids: TypeTextTokenIds,
        attention_mask: TypeTextAttentionMask | None = None,
        *,
        pool: bool = True,
        normalize: bool = True,
    ) -> TypeTextEmbeddings:
        return self.text_encoder(
            token_ids,
            attention_mask,
            pool=pool,
            normalize=normalize,
        )

    def compute_similarities(
        self,
        image_embeddings: TypePooledEmbeddings | TypePatchEmbeddings,
        text_embeddings: TypePooledEmbeddings,
    ) -> TypePatchLogits | TypePooledLogits:
        text_embeddings = rearrange(text_embeddings, "num_prompts c -> c num_prompts")
        is_grid = image_embeddings.ndim == 5
        if is_grid:
            num_images, _, x, y, z = image_embeddings.shape
            image_embeddings = rearrange(
                image_embeddings,
                "num_images c x y z -> (num_images x y z) c",
            )
            similarities_flat = image_embeddings @ text_embeddings
            similarities = rearrange(
                similarities_flat,
                "(num_images x y z) num_prompts -> num_images num_prompts x y z",
                num_images=num_images,
                x=x,
                y=y,
                z=z,
            )
        else:
            similarities = image_embeddings @ text_embeddings
        return similarities

    def classify(
        self,
        images: TypePooledEmbeddings | TypePatchEmbeddings,
        text: TypePooledEmbeddings,
    ) -> TypePooledProbabilities | TypePatchProbabilities:
        logits = self.compute_similarities(images, text)
        assert isinstance(self.softmax_temperature, torch.Tensor)
        probabilities = (logits / self.softmax_temperature).softmax(dim=-1)
        return probabilities

    def save_weights(self, path: TypePath) -> None:
        path = Path(path)
        if path.suffix == ".safetensors":
            save_model(self, str(path))
        else:
            weights = self.state_dict()
            torch.save(weights, path)

    def load_weights(
        self,
        path: TypePath,
        device: torch.device | str | None = None,
    ) -> None:
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Checkpoint file {path} does not exist.")
        if device is None:
            device = torch.device("cpu")
        if path.suffix == ".safetensors":
            if device is not None:
                device = str(device)
            missing, unexpected = load_model(self, path, device=device)
            if missing or unexpected:
                raise RuntimeError("TODO")
        else:
            weights = torch.load(path, map_location=device)
            self.load_state_dict(weights, strict=False)



from __future__ import annotations

import torch.nn.functional as F
from torch import nn
from transformers import BertModel

from ..types import TypeTextAttentionMask
from ..types import TypeTextEmbeddings
from ..types import TypeTextTokenIds
from .base import ModelMixin


class TextEncoder(nn.Module, ModelMixin):
    def __init__(self, backbone: nn.Module, pooler: nn.Module):
        super().__init__()
        if hasattr(backbone, "bert"):
            bert = backbone.bert
            assert isinstance(bert, BertModel)
            bert.pooler = None  # we use our own pooler
            backbone = bert
        self.backbone = backbone
        self.pooler = pooler

    def forward(
        self,
        token_ids: TypeTextTokenIds,
        attention_mask: TypeTextAttentionMask | None = None,
        *,
        pool: bool = True,
        normalize: bool = True,
    ) -> TypeTextEmbeddings:
        token_ids = token_ids.to(self.device)
        if attention_mask is not None:
            attention_mask = attention_mask.to(self.device)
        # We didn't use a projector during pre-training
        output = self.backbone(token_ids, attention_mask=attention_mask)
        embeddings = output["last_hidden_state"]
        if pool:
            embeddings = self.pooler(embeddings)
        if normalize:
            embeddings = F.normalize(embeddings, dim=1)
        return embeddings



from .model.multimodal import get_model
from .pipelines import ZeroShotImageClassificationPipeline
from .processor import get_processor
from .sample_data import load_sample_ct

__all__ = [
    "get_model",
    "get_processor",
    "load_sample_ct",
    "ZeroShotImageClassificationPipeline",
]

__version__ = "0.1.1"




import os
from typing import TypeAlias

import torchio as tio
from jaxtyping import Float
from jaxtyping import Int64
from torch import Tensor

TypePath: TypeAlias = str | os.PathLike[str]
TypePaths: TypeAlias = list[TypePath]

# Raw inputs
TypeImage: TypeAlias = tio.ScalarImage
TypeImages: TypeAlias = list[tio.ScalarImage]
TypeImageOrImages: TypeAlias = TypeImage | TypeImages

TypeString: TypeAlias = str
TypeStrings: TypeAlias = list[str]
TypeStringOrStrings: TypeAlias = TypeString | TypeStrings

TypeIntOrTripletInt: TypeAlias = int | tuple[int, int, int]

# Processed inputs
TypeImagesTensor: TypeAlias = Float[Tensor, "batch 1 x_in y_in z_in"]

TypeTextTokenIds: TypeAlias = Int64[Tensor, "batch text_length"]
TypeTextAttentionMask: TypeAlias = Int64[Tensor, "batch text_length"]

# Outputs
TypeSequenceEmbeddings: TypeAlias = Float[Tensor, "batch sequence_length embed_dim"]
TypePooledEmbeddings: TypeAlias = Float[Tensor, "batch embed_dim"]

TypePatchEmbeddings: TypeAlias = Float[Tensor, "batch embed_dim x_out y_out z_out"]
TypeImageEmbeddings: TypeAlias = TypePatchEmbeddings | TypePooledEmbeddings

TypeTextEmbeddings: TypeAlias = TypeSequenceEmbeddings | TypePooledEmbeddings

TypePatchLogits: TypeAlias = Float[Tensor, "num_images num_prompts x_out y_out z_out"]
TypePooledLogits: TypeAlias = Float[Tensor, "num_images num_prompts"]
TypePatchProbabilities: TypeAlias = TypePatchLogits
TypePooledProbabilities: TypeAlias = TypePooledLogits
TypeScores: TypeAlias = list[dict[str, float | str]]



import torch
from einops import rearrange
from torch import nn

from .types import TypePooledEmbeddings
from .types import TypeSequenceEmbeddings


class AttentionPool1D(nn.Module):
    def __init__(self, embed_dim: int, num_heads: int):
        super().__init__()
        self.attn = nn.MultiheadAttention(embed_dim, num_heads, batch_first=True)

    def forward(self, x: TypeSequenceEmbeddings) -> TypePooledEmbeddings:
        query = x.mean(dim=1, keepdim=True)
        key = value = x
        pooled, _ = self.attn(query, key, value)
        return rearrange(pooled, "batch 1 embed_dim -> batch embed_dim")

    def to_dense(self):
        v_proj_in_weight_qkv = self.get_parameter("attn.in_proj_weight")
        v_proj_in_bias_qkv = self.get_parameter("attn.in_proj_bias")
        v_proj_out_weight = self.get_parameter("attn.out_proj.weight")
        v_proj_out_bias = self.get_parameter("attn.out_proj.bias")
        dim = v_proj_in_weight_qkv.shape[0] // 3
        v_proj_in_weight_v = v_proj_in_weight_qkv[2 * dim :]
        v_proj_in_bias_v = v_proj_in_bias_qkv[2 * dim :]

        value_projection = nn.Conv3d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=1,
        )
        value_projection.weight.data = rearrange(
            v_proj_in_weight_v,
            "c_out c_in -> c_out c_in 1 1 1",
        )
        assert value_projection.bias is not None
        value_projection.bias.data = v_proj_in_bias_v

        out_projection = nn.Conv3d(
            in_channels=dim,
            out_channels=dim,
            kernel_size=1,
        )
        out_projection.weight.data = rearrange(
            v_proj_out_weight,
            "c_out c_in -> c_out c_in 1 1 1",
        )
        assert out_projection.bias is not None
        out_projection.bias.data = v_proj_out_bias

        return nn.Sequential(
            value_projection,
            out_projection,
        )


class MultiLearnedQueryAttentionPool1D(AttentionPool1D):
    def __init__(self, embed_dim: int, num_heads: int):
        super().__init__(embed_dim, num_heads)
        # 4 Queries instead of 1
        self.query = nn.Parameter(torch.randn(1, 4, embed_dim) / embed_dim**0.5)

    def forward(self, x: TypeSequenceEmbeddings) -> TypePooledEmbeddings:
        """
        x: [B, T, D] — sequence of token embeddings
        returns: [B, D] — pooled representation
        """
        B, T, D = x.shape
        query = self.query.expand(B, -1, -1)  # [B, 4, D]
        pooled, _ = self.attn(query, x, x)  # [B, 4, D]
        # pooled: [4, B, D_out], want [B, D_out] by pooling over queries (mean)
        return pooled.mean(dim=1)  # [B, D]



from pathlib import Path

import torchio as tio


def load_sample_ct(clip: bool = True) -> tio.ScalarImage:
    """Get a sample CT image."""
    ct = get_sample_ct()
    if clip:
        clamp = tio.Clamp(-1000, 1000)
        ct = clamp(ct)
    else:
        ct.load()
    return ct


def get_sample_ct() -> tio.ScalarImage:
    """Download a sample CT image if not already present."""
    return tio.datasets.Slicer("CTChest").CT_chest  # type: ignore[reportAttributeAccessIssue]


def get_sample_ct_path() -> Path:
    """Get the path to a sample CT image."""
    return get_sample_ct().path  # type: ignore[reportReturnType]



from __future__ import annotations

import torch
import torch.nn.functional as F
import torchio as tio
from dynamic_network_architectures.architectures.primus import Primus
from einops import rearrange
from torch import nn
from tqdm.auto import tqdm

from ..pooling import AttentionPool1D
from ..types import TypeImageEmbeddings
from ..types import TypeImagesTensor
from ..types import TypeIntOrTripletInt
from .base import ModelMixin


class ImageEncoder(nn.Module, ModelMixin):
    def __init__(
        self,
        backbone: Primus,
        projector: nn.Conv3d,
        pooler: AttentionPool1D,
    ):
        super().__init__()
        self.backbone = self._remove_decoder(backbone)
        self.projector = projector
        self.pooler = pooler

    def _remove_decoder(self, backbone: Primus) -> Primus:
        if hasattr(backbone, "up_projection"):
            backbone.up_projection = nn.Identity()  # type: ignore
        return backbone

    @property
    def patch_size(self) -> tuple[int, int, int]:
        patch_embedder: nn.Conv3d = self.backbone.down_projection.proj  # type: ignore[reportAssignmentType]
        patch_size: tuple[int, int, int] = patch_embedder.stride  # type: ignore[reportAssignmentType]
        return patch_size

    def encode(
        self,
        images: TypeImagesTensor,
    ) -> TypeImageEmbeddings:
        images = images.to(self.device)
        embeddings: TypeImageEmbeddings = self.backbone(images)
        return embeddings

    def encode_sliding_window(
        self,
        images: TypeImagesTensor,
        window_size: TypeIntOrTripletInt,
        overlap: int = 0,
    ) -> TypeImageEmbeddings:
        if len(set(self.patch_size)) > 1:
            msg = (
                "Sliding window encoding is only supported for models with cubic"
                " patch sizes for now."
            )
            raise NotImplementedError(msg)
        else:
            patch_size = self.patch_size[0]
        image_key = "image"  # could be anything
        embeddings = []
        for image in images:
            grid_sampler = tio.inference.GridSampler(
                tio.Subject(**{image_key: tio.ScalarImage(tensor=image)}),
                window_size,
                overlap,
            )
            patch_loader = tio.SubjectsLoader(grid_sampler)  # type: ignore[reportArgumentType]
            aggregator = tio.data.GridAggregator(
                grid_sampler,
                downsampling_factor=patch_size,
            )
            for patches_batch in tqdm(patch_loader):
                input_tensor = patches_batch[image_key][tio.DATA].to(self.device)
                locations = patches_batch[tio.LOCATION]
                outputs = self.backbone(input_tensor)
                aggregator.add_batch(outputs, locations)
            embeddings.append(aggregator.get_output_tensor())
        return torch.stack(embeddings).to(self.device)

    def forward(
        self,
        images: TypeImagesTensor,
        *,
        project: bool,
        pool: bool,
        normalize: bool,
        window_size: TypeIntOrTripletInt | None = None,
    ) -> TypeImageEmbeddings:
        if pool and not project:
            msg = "Pooling requires projection to be enabled. Set project=True."
            raise NotImplementedError(msg)
        if window_size is None:
            embeddings = self.encode(images)
        else:
            embeddings = self.encode_sliding_window(images, window_size)
        if project:
            embeddings = self.projector(embeddings)
            if pool:
                sequence = rearrange(embeddings, "b c x y z -> b (x y z) c")
                embeddings = self.pooler(sequence)
            else:
                embeddings = self.pooler.to_dense()(embeddings)
        if normalize:
            embeddings = F.normalize(embeddings, dim=1)
        return embeddings



import importlib.resources
from pathlib import Path

package_name = __package__
assert isinstance(package_name, str)

REPO_USER = "microsoft"
REPO_NAME = package_name
REVISION = None
WEIGHTS_FILENAME = "model.safetensors"
ROOT_CONFIG_FILENAME = "config.yaml"


def get_configs_dir(pkg: str | None = None) -> Path:
    package_dir = importlib.resources.files(package_name)
    return Path(package_dir / "configs")


__all__ = [
    "get_configs_dir",
]



from pathlib import Path

from huggingface_hub import hf_hub_download
from omegaconf import DictConfig
from omegaconf import OmegaConf
from torch import Tensor

from .defaults import REPO_NAME
from .defaults import REPO_USER
from .defaults import REVISION
from .defaults import ROOT_CONFIG_FILENAME
from .defaults import WEIGHTS_FILENAME
from .defaults import get_configs_dir

TypeStateDict = dict[str, Tensor]


def download_weights(**kwargs) -> Path:
    weights_path = _download_from_hugging_face(**kwargs)
    return weights_path


def _download_from_hugging_face(
    repo_user: str = REPO_USER,
    repo_name: str = REPO_NAME,
    revision: str | None = REVISION,
    filename: str = WEIGHTS_FILENAME,
) -> Path:
    repo_id = f"{repo_user}/{repo_name}"
    try:
        weights_path = hf_hub_download(
            repo_id=repo_id,
            filename=filename,
            revision=revision,
        )
    except Exception as e:
        msg = f'Failed to download "{filename}" from Hugging Face Hub repo "{repo_id}".'
        raise RuntimeError(msg) from e
    return Path(weights_path)


def _resolve_defaults(config_dir: Path, config_name: str) -> DictConfig:
    """Load a YAML config and recursively resolve its Hydra-style defaults."""
    config_path = config_dir / f"{config_name}.yaml"
    raw = OmegaConf.load(config_path)
    assert isinstance(raw, DictConfig)

    defaults = OmegaConf.to_container(raw.pop("defaults", OmegaConf.create([])))
    assert isinstance(defaults, list)

    merged = OmegaConf.create()
    for default in defaults:
        if default == "_self_":
            continue
        if isinstance(default, dict):
            for group, option in default.items():
                sub = _resolve_defaults(config_dir / group, str(option))
                merged = OmegaConf.merge(merged, OmegaConf.create({group: sub}))

    # _self_ is always applied last (standard Hydra behavior)
    merged = OmegaConf.merge(merged, raw)
    assert isinstance(merged, DictConfig)
    return merged


def _load_config(*, overrides: list[str] | None = None) -> DictConfig:
    config_name = ROOT_CONFIG_FILENAME.removesuffix(".yaml")
    config = _resolve_defaults(get_configs_dir(), config_name)
    if overrides:
        config = OmegaConf.merge(config, OmegaConf.from_dotlist(overrides))
    assert isinstance(config, DictConfig)
    return config


def load_model_config(**kwargs) -> DictConfig:
    return _load_config(**kwargs)["model"]


def load_processor_config(**kwargs) -> DictConfig:
    return _load_config(**kwargs)["processor"]
